(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[324],{7614:function(n,e,u){(window.__NEXT_P=window.__NEXT_P||[]).push(["/workspaces/[name]",function(){return u(9431)}])},9431:function(n,e,u){"use strict";u.r(e),u.d(e,{default:function(){return s}});var r=u(5893);u(7294);var t=u(1163),i=u(5843);function s(){let n=(0,t.useRouter)(),{name:e}=n.query;return n.isReady&&e?(0,r.jsx)(i.I,{workspaceName:e}):(0,r.jsx)("div",{children:"Loading..."})}}},function(n){n.O(0,[146,573,488,320,470,990,627,236,843,888,774,179],function(){return n(n.s=7614)}),_N_E=n.O()}]);